#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json
import os, sys
import time
import os.path
import gettext
import common_utils
from ctypes import *
from os.path import abspath, basename, join, isfile, isdir


class HALicenseInfo(Structure):
    """Equivalence of C Structure containing HALicense Information"""
    _fields_ = [("version_id", c_int), ("register_time", c_ulong),
                ("valid_period", c_int), ("feedback", c_char_p)]


libpath = "/usr/lib64/libhalicense.so"
if os.path.exists(libpath):
    haliclib = CDLL(libpath)
    halicense_init = haliclib.ha_license_init
    halicense_init.restype = POINTER(HALicenseInfo)
    halicense_get_fbcode = haliclib.ha_license_get_feedback
    halicense_get_fbcode.restype = c_char_p
    halicense_is_expired = haliclib.halicense_is_expired


def get_ha_host_license():
    license_info = halicense_init()
    # 该字段标识license是否被激活，比如说更换网卡
    is_active = True
    ret = haliclib.halicense_get_info(license_info)
    if ret < 0:
        is_active = False

    time_format = '%Y-%m-%d %H:%M:%S'  # eg. 2012-07-31
    register_text = time.strftime(
        time_format, time.localtime(license_info.contents.register_time))
    valid_text = time.strftime(
        time_format,
        time.localtime(license_info.contents.register_time +
                       license_info.contents.valid_period * 24 * 3600))

    now_time = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))
    halicense_expired = halicense_is_expired()
    license_dic = {
        'version_id': license_info.contents.version_id,
        'register_time': register_text,
        'valid_period': valid_text,
        'now_time': now_time,
        'halicense_is_expired': halicense_expired,
        'is_active': is_active
    }
    feedback_code = halicense_get_fbcode()
    license_dic['feedback'] = str(feedback_code, encoding='utf-8')
    return {'action': True, 'data': license_dic}


def get_remote_ha_host_license(nodeauth):
    if 'nodeauth' in nodeauth:
        node_lic = {}
        for nodeinfo in nodeauth['nodeauth']:

            nodeid = nodeinfo['nodeid']
            # 通过查找配置文件获取nodeid所在主机上的所有ip
            nodeips = node.get_node_ip(nodeid)
            if 'data' in nodeips and 'ips' in nodeips[
                    'data'] and nodeips['data']['ips'] != []:
                nodeip = nodeips['data']['ips'][0]
            else:
                return {
                    'action': False,
                    'info': str(nodeid) + _(": Node needs to configure IP")
                }
            password = nodeinfo['password']
            remote = Remote(nodeip, password)
            try:
                data = {}
                results = remote.get_license_info()
                if nodeip in results and results[nodeip] != "":
                    info = eval(results[nodeip])
                    if 'data' in info:
                        data = info['data']
                node_lic[nodeid] = data

            # fabric 的异常比较特殊， 异常时SystemExit
            except SystemExit:
                return {
                    'action': False,
                    'error': _('Get remote license info failed')
                }

        return {'action': True, 'data': node_lic}

    return {'action': False, 'error': _('Node need auth info')}


def install_ha_host_license_own(fileinfo):
    if fileinfo is None:
        return {'action': False, 'error': _('Install license failed')}
    buf = fileinfo['file']
    filename = '/etc/corosync/license.dat.new'
    outfile = open(filename, 'wb')
    outfile.write(bytes(buf, encoding='utf-8'))
    outfile.close()
    cmd = "halicense_install " + filename
    status, output = common_utils.run_cmd(cmd)
    if status != 0:
        return {'action': False, 'error': _('Install license failed')}
    else:
        ret = get_ha_host_license()
        if ret['action'] == True:
            ret['info'] = _('Install license success')
        return {'action': True, 'data': ret}



if __name__ == '__main__':
    '''
    data =   {
        "nodeauth":[
        {
            "nodeid":"ns187",
            "password":"qwer1234"
        },
        {
            "nodeid":"ns188",
            "password":"qwer1234"
        }
        ]
        }
    '''
    # print get_remote_ha_host_license(data)
    # 这行必须存在是远程获取license命令
    print(get_ha_host_license())
